/*******************************************************************************

  Intel Data Center Bridging (DCB) Software
  Copyright(c) 2007-2008 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Contact Information:
  e1000-eedc Mailing List <e1000-eedc@lists.sourceforge.net>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include "dcb_osdep.h"
#include "ports.h"
#include "l2_packet.h"
#include "tlv.h"
#include "states.h"
#include "mibdata.h"
#include "dcb_protocol.h"
#include "messages.h"

static u32 check_feature_not_present(char *device_name, u32 subtype,
				u32 EventFlag, u32 feature);

void rxInitializeLLDP(struct port *port) {

	port->portrx.rcvFrame = FALSE;
	port->portrx.badFrame = FALSE;
	port->portrx.tooManyNeighbors = FALSE;
	port->portrx.rxInfoAge = FALSE;
	if (port->portrx.framein) {
		free(port->portrx.framein);
		port->portrx.framein = NULL;
	}
	port->portrx.sizein = 0;

	mibDeleteObjects(port);
	return;
}

void rxReceiveFrame(void *ctx, unsigned int ifindex, const u8 *buf, size_t len)
{

	struct port * port;
	u8  frame_error = 0;
	struct l2_ethhdr *hdr;
	struct l2_ethhdr example_hdr,*ex;

	port = (struct port *)ctx;

	if (port->portrx.framein)
		free(port->portrx.framein);

	port->portrx.framein = (u8 *)malloc(len);
	if (port->portrx.framein == NULL) {
		printf("ERROR - could not allocate memory for rx'ed frame\n");
		return;
	}
	memcpy(port->portrx.framein, buf, len);

	port->portrx.sizein = (u16)len;
	ex = &example_hdr;
	memcpy(ex->h_dest, multi_cast_source, ETH_ALEN);
	ex->h_proto = htons(ETH_P_LLDP);
	hdr = (struct l2_ethhdr *)port->portrx.framein;

	if ((memcmp(hdr->h_dest,ex->h_dest, ETH_ALEN) != 0)) {
		printf("ERROR LLDP multicast address error in incoming frame. "
			"Dropping frame.\n");
		frame_error++;
		free(port->portrx.framein);
		port->portrx.framein = NULL;
		port->portrx.sizein = 0;
		return;
	}

	if (hdr->h_proto != example_hdr.h_proto) {
		printf("ERROR Ethertype not LLDP ethertype but ethertype "
			"'%x' in incoming frame.\n", htons(hdr->h_proto));
		frame_error++;
		free(port->portrx.framein);
		port->portrx.framein = NULL;
		port->portrx.sizein = 0;
		return;
	}

	if (!frame_error) {
		port->portstats.statsFramesInTotal++;
		port->portrx.rcvFrame = 1;
	}
	run_rx_sm(port, FALSE);
}

void rxProcessFrame(struct port * port)
{
	u16 tlv_cnt = 0;
	u8  tlv_type = 0;
	u16 tlv_length = 0;
	u16 tlv_offset = 0;
	u16 *tlv_head_ptr = NULL;
	u8  frame_error = 0;
	u8  oui[DCB_OUI_LEN] = {0x00,0x1b,0x21};
	boolean_t msap_compare_1 = FALSE;
	boolean_t msap_compare_2 = FALSE;
	boolean_t good_neighbor  = FALSE;
	boolean_t tlv_stored     = FALSE;

	assert(port->portrx.framein && port->portrx.sizein);
	port->lldpdu = 0;
	port->dcbdu = 0;

	port->portrx.rxmanifest = (manifest *)malloc(sizeof(manifest));
	if (port->portrx.rxmanifest == NULL) {
		printf("ERROR - could not allocate memory for receive "
			"manifest\n");
		return;
	}
	memset(port->portrx.rxmanifest,0, sizeof(manifest));

	tlv_offset = sizeof(struct l2_ethhdr);  /* Points to 1st TLV */

	do {
		tlv_cnt++;
		if (tlv_offset > port->portrx.sizein) {
			printf("ERROR: Frame overrun!\n");
			frame_error++;
			goto out;
		}

		tlv_head_ptr = (u16 *)&port->portrx.framein[tlv_offset];
		tlv_length = htons(*tlv_head_ptr) & 0x01FF;
		tlv_type = (u8)(htons(*tlv_head_ptr) >> 9);

		if (tlv_cnt <= 3) {
			if (tlv_cnt != tlv_type) {
				printf("ERROR:TLV missing or TLVs out "
					"of order!\n");
				frame_error++;
				goto out;
			}
		}

		if (tlv_cnt > 3) {
			if ((tlv_type == 1) || (tlv_type == 2) ||
				(tlv_type == 3)) {
				printf("ERROR: Extra Type 1 Type2, or "
					"Type 3 TLV!\n");
				frame_error++;
				goto out;
			}
		}

		if ((tlv_type == TIME_TO_LIVE_TLV) && (tlv_length != 2)) {
			printf("ERROR:TTL TLV validation error! \n");
			frame_error++;
			goto out;
		}

		u16 tmp_offset = tlv_offset + tlv_length;
		if (tmp_offset > port->portrx.sizein) {
			printf("ERROR: Frame overflow error: offset=%d, "
				"rx.size=%d \n", tmp_offset,
				port->portrx.sizein);
			frame_error++;
			goto out;
		}

		u8 *info_string = (u8 *)&port->portrx.framein[tlv_offset +
							sizeof(*tlv_head_ptr)];

		struct unpacked_tlv *tlv = create_tlv();

		if (!tlv) {
			printf("ERROR: Failed to malloc space for "
				"incoming TLV. \n");
			goto out;
		}

		tlv->type = tlv_type;
		tlv->length = tlv_length;
		tlv->info_string = (u8 *)malloc(tlv_length);
		if (tlv->info_string) {
			memset(tlv->info_string,0, tlv_length);
			memcpy(tlv->info_string, info_string, tlv_length);
		} else {
			printf("ERROR: Failed to malloc space for incoming "
				"TLV info_string \n");
			destruct_unpacked_tlv(tlv);
			goto out;
		}

		/* Validate the TLV */
		tlv_offset += sizeof(*tlv_head_ptr) + tlv_length;
		/* Get MSAP info */
		if (tlv->type == TYPE_1) {
			if (port->lldpdu & RECEIVED_LLDP_TLV_TYPE1) {
				printf("Received multiple Type 1 TLVs in this "
					"LLDPDU\n");
				frame_error++;
				destruct_unpacked_tlv(tlv);
				goto out;
			} else {
				port->lldpdu |= RECEIVED_LLDP_TLV_TYPE1;
				port->portrx.rxmanifest->Type1 = tlv;
				tlv_stored = TRUE;
			}

			if (port->msap.msap1 == NULL) {
				port->msap.length1 = tlv->length;
				port->msap.msap1 = (u8 *)malloc(tlv->length);
				if (!(port->msap.msap1)) {
					printf("ERROR: Failed to malloc space "
						"for msap1 \n");
					goto out;
				}
				memcpy(port->msap.msap1, tlv->info_string,
					tlv->length);
			} else {
				if (tlv->length == port->msap.length1) {
					if ((memcmp(tlv->info_string,
						port->msap.msap1,
						tlv->length) == 0))
						msap_compare_1 = TRUE;
				}
			}
		}
		if (tlv->type == TYPE_2) {
			if (port->lldpdu & RECEIVED_LLDP_TLV_TYPE2) {
				printf("Received multiple Type 2 TLVs in this "
					"LLDPDU\n");
				frame_error++;
				destruct_unpacked_tlv(tlv);
				goto out;
			} else {
				port->lldpdu |= RECEIVED_LLDP_TLV_TYPE2;
				port->portrx.rxmanifest->Type2 = tlv;
				tlv_stored = TRUE;
			}

			if (port->msap.msap2 == NULL) {
				port->msap.length2 = tlv->length;
				port->msap.msap2 = (u8 *)malloc(tlv->length);
				if (!(port->msap.msap2)) {
					printf("ERROR: Failed to malloc space "
						"for msap2 \n");
					goto out;
				}
				memcpy(port->msap.msap2, tlv->info_string,
					tlv->length);
			} else {
				if (tlv->length == port->msap.length2) {
					if ((memcmp(tlv->info_string,
						port->msap.msap2,
						tlv->length) == 0))
						msap_compare_2 = TRUE;
				}
				if ((msap_compare_1 == TRUE) &&
					(msap_compare_2 == TRUE)) {
					msap_compare_1 = FALSE;
					msap_compare_2 = FALSE;
					good_neighbor = TRUE;
				} else {
					/* New neighbor */
					port->portrx.tooManyNeighbors = TRUE;
				}
			}
		}
		if (tlv->type == TYPE_3) {
			if (port->lldpdu & RECEIVED_LLDP_TLV_TYPE3) {
				printf("Received multiple Type 3 TLVs in this "
					"LLDPDU\n");
				frame_error++;
				destruct_unpacked_tlv(tlv);
				goto out;
			} else {
				port->lldpdu |= RECEIVED_LLDP_TLV_TYPE3;
				port->portrx.rxmanifest->Type3 = tlv;
				tlv_stored = TRUE;
			}
			if ((port->portrx.tooManyNeighbors == TRUE) &&
				(good_neighbor == FALSE)) {
				printf("tooManyNeighbors, no good neighbor\n");
				port->porttimers.tooManyNeighborsTimer =
					ntohs(*(u16 *)tlv->info_string);
				msap_compare_1 = FALSE;
				msap_compare_2 = FALSE;
				goto out;
			} else {
				port->porttimers.rxTTL =
					ntohs(*(u16 *)tlv->info_string);
				good_neighbor = FALSE;
			}
		}
		if (tlv->type == TYPE_4) {
			port->lldpdu |= RECEIVED_LLDP_TLV_TYPE4;
					port->portrx.rxmanifest->Type4 = tlv;
			tlv_stored = TRUE;
		}
		if (tlv->type == TYPE_5) {
			port->lldpdu |= RECEIVED_LLDP_TLV_TYPE5;
					port->portrx.rxmanifest->Type5 = tlv;
			tlv_stored = TRUE;
		}
		if (tlv->type == TYPE_6) {
			port->lldpdu |= RECEIVED_LLDP_TLV_TYPE6;
			port->portrx.rxmanifest->Type6 = tlv;
			tlv_stored = TRUE;
		}
		if (tlv->type == TYPE_7) {
			port->lldpdu |= RECEIVED_LLDP_TLV_TYPE7;
					port->portrx.rxmanifest->Type7 = tlv;
			tlv_stored = TRUE;
		}
		if (tlv->type == TYPE_8) {
			port->lldpdu |= RECEIVED_LLDP_TLV_TYPE8;
					port->portrx.rxmanifest->Type8 = tlv;
			tlv_stored = TRUE;
		}
		/* Process a Type127 to see if it is a DCB TLV */
		if (tlv->type == TYPE_127) {
			if (port->lldpdu & RECEIVED_LLDP_TLV_TYPE127_DCB_TLV) {
				printf("Received multiple Type 127 TLVs in "
					"this LLDPDU\n");
				frame_error++;
				destruct_unpacked_tlv(tlv);
				goto out;
			} 
			/* Is a DCB TLV */
			if (tlv->length < (DCB_OUI_LEN + OUI_SUBTYPE_LEN)) {
				printf("Received invalid Type 127 TLV\n");
				frame_error++;
				destruct_unpacked_tlv(tlv);
					goto out;
			}
			if ((memcmp(tlv->info_string, &oui, DCB_OUI_LEN) == 0)) {
				assert(port->porttlvs.current_peer_tlv
					== NULL);
				port->lldpdu |=
					RECEIVED_LLDP_TLV_TYPE127_DCB_TLV;
				port->portrx.rxmanifest->Type127 = tlv;
				tlv_stored = TRUE;
				if (unpack_dcbtlvs(port,
					port->portrx.rxmanifest->Type127) ==
					FALSE) {
					printf("Had an error unpacking "
						"the DCB TLVs Trashing "
						"LLDPDU\n");
					frame_error++;
					goto out;
				}
				load_peer_tlvs(port,tlv, CURRENT_PEER);
			}
		}
		if (tlv->type == TYPE_0) {
			destruct_unpacked_tlv(tlv);
			tlv_stored = TRUE;
			if (port->porttlvs.current_peer_tlv) {
				process_dcb_tlv(port,
					port->porttlvs.current_peer_tlv);
			}
		}

		if (!tlv_stored) {
			printf("\nrxProcessFrame: allocated TLV "
				   " was not stored! (%p)\n", tlv);
			assert(tlv_stored);
		}
		tlv = NULL;
	} while(tlv_type != 0);

out:
	if (frame_error) {
		/* Trashing the frame because of errors. */
		port->portstats.statsFramesDiscardedTotal++;
		port->portstats.statsFramesInErrorsTotal++;
		port->portrx.badFrame = TRUE;
		/* return; */
	}
	/* Maybe want to zero out the frame buffer on exit?
	 * bzero(port->portrx.framein, port->portrx.sizein);
	 */
	free(port->portrx.framein);
	port->portrx.framein = NULL;
	port->portrx.sizein = 0;
	port->lldpdu = 0;
	port->dcbdu = 0;
	if (port->porttlvs.current_peer_tlv)
		port->porttlvs.current_peer_tlv =
			destruct_unpacked_tlv(port->porttlvs.current_peer_tlv);
	clear_manifest(port);

	return;
}

u8 mibDeleteObjects(struct port *port)
{
	control_protocol_attribs  peer_control_setting;
	pg_attribs  peer_pg_setting;
	pfc_attribs peer_pfc_setting;
	app_attribs peer_app_setting;
	u32 subtype = 0;
	u32 EventFlag = 0;

	/* Set any stored values for this TLV to !Present */
	if (get_peer_pg(port->ifname, &peer_pg_setting) == dcb_success) {
		if (peer_pg_setting.protocol.RxFeatureTLVPresent == TRUE) {
			peer_pg_setting.protocol.RxFeatureTLVPresent = FALSE;
			put_peer_pg(port->ifname, &peer_pg_setting);
			DCB_SET_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PG);
		}
	} else {
		return (u8)-1;
	}

	if (get_peer_pfc(port->ifname, &peer_pfc_setting) == dcb_success) {
		if (peer_pfc_setting.protocol.RxFeatureTLVPresent == TRUE) {
			peer_pfc_setting.protocol.RxFeatureTLVPresent = FALSE;
			put_peer_pfc(port->ifname, &peer_pfc_setting);
			DCB_SET_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PFC);
		}
	} else {
		return (u8)-1;
	}

	if (get_peer_app(port->ifname, subtype, &peer_app_setting) ==
		dcb_success) {
		if (peer_app_setting.protocol.RxFeatureTLVPresent == TRUE) {
			peer_app_setting.protocol.RxFeatureTLVPresent = FALSE;
			peer_app_setting.Length = 0;
			put_peer_app(port->ifname, subtype, &peer_app_setting);
			DCB_SET_FLAGS(EventFlag, DCB_REMOTE_CHANGE_APPTLV);
		}
	} else {
		return (u8)-1;
	}
	if (get_peer_control(port->ifname, &peer_control_setting) ==
		dcb_success) {
		peer_control_setting.RxDCBTLVState = DCB_PEER_EXPIRED;
		put_peer_control(port->ifname, &peer_control_setting);
	} else {
		return (u8)-1;
	}

	peer_control_setting.RxDCBTLVState = DCB_PEER_PRESENT;
	if (EventFlag != 0) {
		/* process for all subtypes */
		run_dcb_protocol(port->ifname, EventFlag, DCB_MAX_APPTLV+1);
		EventFlag = 0;
	}

	/* Clear history */
	if (port->porttlvs.last_peer_tlv) {
		port->porttlvs.last_peer_tlv =
			destruct_unpacked_tlv(port->porttlvs.last_peer_tlv);
	}

	port->msap.length1 = 0;
	if (port->msap.msap1) {
		free(port->msap.msap1);
		port->msap.msap1 = NULL;
	}

	port->msap.length2 = 0;
	if (port->msap.msap2) {
		free(port->msap.msap2);
		port->msap.msap2 = NULL;
	}
	return 0;
}

/* for the specified remote feature, if the feature is not present in the
 * EventFlag parameter (indicating it was not received in the DCB TLV), then
 * check and update the peer data store object for the feature if it is
 * currently marked as being present.
 *
 * returns TRUE if the feature is not present now
 *              the peer data store feature object is set to 'not present'
 *         FALSE otherwise.
*/
static u32 check_feature_not_present(char *device_name, u32 subtype,
				u32 EventFlag, u32 feature)
{
	pg_attribs   peer_pg_setting;
	pfc_attribs  peer_pfc_setting;
	app_attribs  peer_app_setting;

	/* if (!DCB_TEST_FLAGS(EventFlag, feature, feature)) { */
	if (DCB_TEST_FLAGS(EventFlag, feature, feature))
		return FALSE;

	switch (feature) {
	case DCB_REMOTE_CHANGE_PG:
		if ((get_peer_pg(device_name, &peer_pg_setting) == dcb_success)
			&& (peer_pg_setting.protocol.RxFeatureTLVPresent ==
			TRUE)) {
			peer_pg_setting.protocol.RxFeatureTLVPresent = FALSE;
			put_peer_pg(device_name, &peer_pg_setting);
		}
		break;
	case DCB_REMOTE_CHANGE_PFC:
		if ((get_peer_pfc(device_name, &peer_pfc_setting) == 
			dcb_success) && 
			(peer_pfc_setting.protocol.RxFeatureTLVPresent == TRUE))
			{
			peer_pfc_setting.protocol.RxFeatureTLVPresent = FALSE;
			put_peer_pfc(device_name, &peer_pfc_setting);
		}
		break;
	case DCB_REMOTE_CHANGE_APPTLV:
		if ((get_peer_app(device_name, subtype, &peer_app_setting) ==
			dcb_success) &&
			(peer_app_setting.protocol.RxFeatureTLVPresent == TRUE))
			{
			peer_app_setting.protocol.RxFeatureTLVPresent = FALSE;
			peer_app_setting.Length = 0;
			put_peer_app(device_name, subtype, &peer_app_setting);
		}
		break;
	default:
		break;
	}

	return TRUE;
}


void  mibUpdateObjects(struct port *port)
{
	u32 EventFlag = 0;

	if (port->portrx.rxmanifest->DCB_Type1) {
		if (process_dcb_type1_tlv(port) != TRUE) {
			/* Error Set error condition for all features
			 * on this port and trash DCB TLV */
		}
	} else {
			/* Error Set error condition for all features
			 * on this port and trash DCB TLV */
	}
	if (port->portrx.rxmanifest->DCB_Type2) {
		if (process_dcb_type2_tlv(port) != TRUE) {
			/*Error Set error condition for this feature */
		} else {
			DCB_SET_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PG);
		}
	} else {
		if (check_feature_not_present(port->ifname, 0,
			EventFlag, DCB_REMOTE_CHANGE_PG)) {
			DCB_SET_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PG);
		}
	}
	if (port->portrx.rxmanifest->DCB_Type3) {
		 if (process_dcb_type3_tlv(port) != TRUE) {
			 /* Error Set error condition for feature */
		 } else {
			 DCB_SET_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PFC);
		 }
	} else {
		if (check_feature_not_present(port->ifname, 0,
			EventFlag, DCB_REMOTE_CHANGE_PFC)) {
			DCB_SET_FLAGS(EventFlag, DCB_REMOTE_CHANGE_PFC);
		}
	}
	if (port->portrx.rxmanifest->DCB_Type5) {
		if (process_dcb_type5_tlv(port) != TRUE) {
			/* Error Set error condition for this feature */
		} else {
			DCB_SET_FLAGS(EventFlag,
				DCB_REMOTE_CHANGE_APPTLV);
		}
	} else {
		if (check_feature_not_present(port->ifname, 0,
			EventFlag, DCB_REMOTE_CHANGE_APPTLV)) {
			DCB_SET_FLAGS(EventFlag,
				DCB_REMOTE_CHANGE_APPTLV);
		}
	}

	/* Run the feature and control protocol for all features and subtypes */
	run_dcb_protocol(port->ifname, EventFlag, DCB_MAX_APPTLV+1);
	EventFlag = 0;
	port->rxChanges = TRUE;
	return;
}

void run_rx_sm(struct port *port, int update_timers)
{
	set_rx_state(port);
	do {
		switch(port->portrx.state) {
		case LLDP_WAIT_PORT_OPERATIONAL:
			process_wait_port_operational(port);
			break;
		case DELETE_AGED_INFO:
			process_delete_aged_info(port);
			break;
		case RX_LLDP_INITIALIZE:
			process_rx_lldp_initialize(port);
			break;
		case RX_WAIT_FOR_FRAME:
			process_wait_for_frame(port);
			break;
		case RX_FRAME:
			process_rx_frame(port);
			break;
		case DELETE_INFO:
			process_delete_info(port);
			break;
		case UPDATE_INFO:
			process_update_info(port);
			break;
		default:
			printf("ERROR: The RX State Machine is broken!\n");
			log_message(MSG_ERR_TX_SM_INVALID, NULL);
		}
	} while (set_rx_state(port) == TRUE);

	if (update_timers == TRUE)
		update_rx_timers(port);
}

boolean_t set_rx_state(struct port *port)
{
	if ((port->portrx.rxInfoAge == FALSE) && (port->portEnabled == FALSE)) {
		printf("set_rx_state: portEnabled==FALSE \n");
		rx_change_state(port, LLDP_WAIT_PORT_OPERATIONAL);
	}

	switch(port->portrx.state) {
	case LLDP_WAIT_PORT_OPERATIONAL:
		if (port->portrx.rxInfoAge == TRUE) {
			rx_change_state(port, DELETE_AGED_INFO);
			return TRUE;
		}
		else if (port->portEnabled == TRUE) {
			rx_change_state(port, RX_LLDP_INITIALIZE);
			return TRUE;
		}
		return FALSE;
	case DELETE_AGED_INFO:
		rx_change_state(port, LLDP_WAIT_PORT_OPERATIONAL);
		return TRUE;
	case RX_LLDP_INITIALIZE:
		if ((port->adminStatus == enabledRxTx) ||
			(port->adminStatus == enabledRxOnly)) {
			rx_change_state(port, RX_WAIT_FOR_FRAME);
			return TRUE;
		}
		return FALSE;
	case RX_WAIT_FOR_FRAME:
		if (port->portrx.rxInfoAge == TRUE) {
			rx_change_state(port, DELETE_INFO);
			return TRUE;
		} else if (port->portrx.rcvFrame == TRUE) {
			rx_change_state(port, RX_FRAME);
			return TRUE;
		}
		return FALSE;
	case DELETE_INFO:
		rx_change_state(port, RX_WAIT_FOR_FRAME);
		return TRUE;
	case RX_FRAME:
		if (port->porttimers.rxTTL == 0) {
			rx_change_state(port, DELETE_INFO);
			return TRUE;
		} else if ((port->porttimers.rxTTL != 0) &&
			(port->rxChanges == TRUE)) {
			rx_change_state(port, UPDATE_INFO);
			return TRUE;
		}
		rx_change_state(port, RX_WAIT_FOR_FRAME);
		return TRUE;
	case UPDATE_INFO:
		rx_change_state(port, RX_WAIT_FOR_FRAME);
		return TRUE;
	default:
		printf("ERROR: The RX State Machine is broken!\n");
		log_message(MSG_ERR_RX_SM_INVALID, NULL);
		return FALSE;
	}
}

void process_wait_port_operational(struct port *port)
{
	/* Maybe could sleep for a sec. ? */
	return;
}

void process_delete_aged_info(struct port *port)
{
	port->portrx.somethingChangedRemote = FALSE;
	mibDeleteObjects(port);
	port->portrx.rxInfoAge = FALSE;
	port->portrx.somethingChangedRemote = TRUE;
	return;
}

void process_rx_lldp_initialize(struct port *port)
{
	rxInitializeLLDP(port);
	port->portrx.rcvFrame = FALSE;
	return;
}

void process_wait_for_frame(struct port *port)
{
	port->portrx.badFrame  = FALSE;
	port->portrx.rxInfoAge = FALSE;
	port->portrx.somethingChangedRemote = FALSE;
	return;
}

void process_rx_frame(struct port *port)
{
	port->rxChanges = FALSE;
	port->portrx.rcvFrame = FALSE;
	rxProcessFrame(port);
	return;
}

void process_delete_info(struct port *port)
{
	mibDeleteObjects(port);
	port->portrx.somethingChangedRemote = TRUE;
	return;
}

void process_update_info(struct port *port)
{
	port->portrx.somethingChangedRemote = TRUE;
	return;
}

void update_rx_timers(struct port *port)
{

	if (port->porttimers.rxTTL) {
		port->porttimers.rxTTL--;
		if (port->porttimers.rxTTL == 0) {
			port->portrx.rxInfoAge = TRUE;
		}
	}
	if (port->porttimers.tooManyNeighborsTimer) {
		port->porttimers.tooManyNeighborsTimer--;
		if (port->porttimers.tooManyNeighborsTimer == 0) {
			port->portrx.tooManyNeighbors = FALSE;
		}
	}
}

void rx_change_state(struct port *port, u8 newstate) {
	switch(newstate) {
		case LLDP_WAIT_PORT_OPERATIONAL:
			break;
		case RX_LLDP_INITIALIZE:
			assert(port->portrx.state ==
				LLDP_WAIT_PORT_OPERATIONAL);
			break;
		case DELETE_AGED_INFO:
			assert(port->portrx.state ==
				LLDP_WAIT_PORT_OPERATIONAL);
			break;
		case RX_WAIT_FOR_FRAME:
			if (!(port->portrx.state == RX_LLDP_INITIALIZE ||
				port->portrx.state == DELETE_INFO ||
				port->portrx.state == UPDATE_INFO ||
				port->portrx.state == RX_FRAME)) {
				assert(port->portrx.state !=
					RX_LLDP_INITIALIZE);
				assert(port->portrx.state != DELETE_INFO);
				assert(port->portrx.state != UPDATE_INFO);
				assert(port->portrx.state != RX_FRAME);
			}
			break;
		case RX_FRAME:
			assert(port->portrx.state == RX_WAIT_FOR_FRAME);
			break;
		case DELETE_INFO:
			if (!(port->portrx.state == RX_WAIT_FOR_FRAME ||
				port->portrx.state == RX_FRAME)) {
				assert(port->portrx.state == RX_WAIT_FOR_FRAME);
				assert(port->portrx.state == RX_FRAME);
			}
			break;
		case UPDATE_INFO:
			assert(port->portrx.state == RX_FRAME);
			break;
		default:
			printf("ERROR: The RX State Machine is broken!\n");
			log_message(MSG_ERR_RX_SM_INVALID, NULL);
	}
	port->portrx.state = newstate;
}


void load_peer_tlvs(struct port *port,struct unpacked_tlv *tlv,int type)
{
	struct unpacked_tlv *tmp = NULL;

	tmp = create_tlv();
	if (tmp) {
		tmp->type = tlv->type;
		tmp->length = tlv->length;
		tmp->info_string = (u8 *)malloc(tlv->length);
		if (tmp->info_string) {
			memset(tmp->info_string,0, tlv->length);
			memcpy(tmp->info_string,tlv->info_string,tlv->length);
		} else {
			printf("load_peer_tlvs: Failed to malloc info_string\n");
		}

		if (type == CURRENT_PEER) {
			assert(port->porttlvs.current_peer_tlv == NULL);
			port->porttlvs.current_peer_tlv = tmp;
		}
		else
		if (type == LAST_PEER) {
			assert(port->porttlvs.last_peer_tlv == NULL);
			port->porttlvs.last_peer_tlv = tmp;
		}
	}
	return;
}

boolean_t  unpack_dcbtlvs(struct port *port, struct unpacked_tlv *tlv)
{
	/* unpack the tlvs and store in manifest */
	u8 *offset = NULL;   /* iterator */
	u16 current = 0, tl = 0;
	u16 end = 0;         /* End of data blob */
	struct unpacked_tlv     *dcbtlv;

	/* OUI + subtype sizes equal the start of data blob */
	offset = (u8  *)&tlv->info_string[OUI_SUBTYPE_LEN + DCB_OUI_LEN];
	end = tlv->length  - (OUI_SUBTYPE_LEN + DCB_OUI_LEN);
	/* Process */
	do {
		dcbtlv = create_tlv();
		if (!dcbtlv) {
			printf("ERROR: Failed to malloc space for incoming "
				"DCB TLV. \n");
			return FALSE;
		}
		memcpy(&tl, offset, sizeof(tl));
		offset += sizeof(tl);
		dcbtlv->length = ntohs(tl) & 0x01FF;
		dcbtlv->type   = (u8)(ntohs(tl) >> 9);
		dcbtlv->info_string = (u8 *)malloc(dcbtlv->length);
		if (dcbtlv->info_string) {
			memset(dcbtlv->info_string, 0, dcbtlv->length);
			memcpy(dcbtlv->info_string, offset, dcbtlv->length);
		} else {
			printf("ERROR: Failed to malloc space for incoming "
				"TLV info_string \n");
			destruct_unpacked_tlv(dcbtlv);
			return FALSE;
		}
		current += dcbtlv->length + sizeof(tl);
		offset += dcbtlv->length;
		switch(dcbtlv->type) {
			case DCB_CONTROL_TLV:
				port->portrx.rxmanifest->DCB_Type1 = dcbtlv;
			break;
			case DCB_PRIORITY_GROUPS_TLV:
				port->portrx.rxmanifest->DCB_Type2 = dcbtlv;
			break;
			case DCB_PRIORITY_FLOW_CONTROL_TLV:
				port->portrx.rxmanifest->DCB_Type3 = dcbtlv;
			break;
			case DCB_APPLICATION_TLV:
				port->portrx.rxmanifest->DCB_Type5 = dcbtlv;
			break;
			default:
				destruct_unpacked_tlv(dcbtlv);
			break;
		}
		dcbtlv = NULL;
	} while(current < end);

	return TRUE;
}

void process_dcb_tlv(struct port *port,struct unpacked_tlv *tlv)
{
	if (port->porttlvs.last_peer_tlv == NULL) {
		/* First tlv. Create and load */
		load_peer_tlvs(port,tlv, LAST_PEER);
		mibUpdateObjects(port); /* Put 1st in peer mem */
	} else {
		if (tlv->length == port->porttlvs.last_peer_tlv->length) {
			if (memcmp(tlv->info_string,
				port->porttlvs.last_peer_tlv->info_string,
				tlv->length) != 0) {
				/* Strings are different unpack the tlvs and
				 * store in peer mem
				 * Load tlv in last_peer_tlv
				*/
				mibUpdateObjects(port);
				port->porttlvs.last_peer_tlv->type = tlv->type;
				port->porttlvs.last_peer_tlv->length =
					tlv->length;
				if (port->porttlvs.last_peer_tlv->info_string) {
				  memcpy(
					port->porttlvs.last_peer_tlv->info_string,
					tlv->info_string,tlv->length);
				}
			}
		} else {
			/* Lengths are different unpack the tlvs and store in
			 * peer memory. Load tlv in last_peer_tlv
			*/
			free(port->porttlvs.last_peer_tlv->info_string);
			free(port->porttlvs.last_peer_tlv);
			port->porttlvs.last_peer_tlv = NULL;
			load_peer_tlvs(port,tlv, LAST_PEER);
			mibUpdateObjects(port);
		}
	}
	port->porttlvs.current_peer_tlv = destruct_unpacked_tlv(tlv);
	return;
}


boolean_t process_dcb_type1_tlv(struct port *port)
{
	control_protocol_attribs  peer_control_setting;

	peer_control_setting.Oper_version =
		port->portrx.rxmanifest->DCB_Type1->info_string
		[DCB_Type1_OPER_VERSION_OFFSET];
	peer_control_setting.Max_version =
		port->portrx.rxmanifest->DCB_Type1->info_string
		[DCB_Type1_MAX_VERSION_OFFSET];

	u32 tmp32 = 0;
	memcpy(&tmp32,
		&port->portrx.rxmanifest->DCB_Type1->info_string
		[DCB_Type1_SEQNO_OFFSET], sizeof(u32));
	peer_control_setting.SeqNo = ntohl(tmp32);
	tmp32 = 0;
	memcpy(&tmp32,
		&port->portrx.rxmanifest->DCB_Type1->info_string
		[DCB_Type1_ACKNO_OFFSET], sizeof(u32));
	peer_control_setting.AckNo = ntohl(tmp32);
	printf("*** Received a DCB_CONTROL_TLV: -- SeqNo=%d, AckNo=%d \n",
		peer_control_setting.SeqNo, peer_control_setting.AckNo);
	peer_control_setting.RxDCBTLVState = DCB_PEER_PRESENT;
	put_peer_control(port->ifname, &peer_control_setting);

	return(TRUE);
}


boolean_t 	process_dcb_type2_tlv(struct port *port)
{
	pg_attribs   peer_pg_setting;
	int i = 0;

	peer_pg_setting.protocol.Advertise = TRUE;
	peer_pg_setting.protocol.Oper_version =
		port->portrx.rxmanifest->DCB_Type2->info_string
		[DCB_HEADER_OPER_VERSION_OFFSET];
	peer_pg_setting.protocol.Max_version =
		port->portrx.rxmanifest->DCB_Type2->info_string
		[DCB_HEADER_MAX_VERSION_OFFSET];
	if (port->portrx.rxmanifest->DCB_Type2->info_string
		[DCB_HEADER_EWE_OFFSET] & BIT7) {
		peer_pg_setting.protocol.Enable = TRUE;
	} else {
		peer_pg_setting.protocol.Enable = FALSE;
	}
	if (port->portrx.rxmanifest->DCB_Type2->info_string
		[DCB_HEADER_EWE_OFFSET] & BIT6) {
		peer_pg_setting.protocol.Willing = TRUE;
	} else {
		peer_pg_setting.protocol.Willing = FALSE;
	}
	if (port->portrx.rxmanifest->DCB_Type2->info_string
		[DCB_HEADER_EWE_OFFSET] & BIT5) {
		peer_pg_setting.protocol.Error = TRUE;
	} else {
		peer_pg_setting.protocol.Error = FALSE;
	}
	peer_pg_setting.protocol.RxFeatureTLVPresent = TRUE;
	for (i = 0; i < 8; i++) {
		peer_pg_setting.tx.bwg_percent_of_link[i] =
			port->portrx.rxmanifest->DCB_Type2->info_string
			[DCB_Type2_PERCENTAGE_OFFSET + i];
		peer_pg_setting.rx.bwg_percent_of_link[i] =
			port->portrx.rxmanifest->DCB_Type2->info_string
			[DCB_Type2_PERCENTAGE_OFFSET + i];

		u8 tmp_bwg_id = port->portrx.rxmanifest->DCB_Type2->info_string
			[DCB_Type2_SETTINGS_OFFSET + 2*i + BYTE1_OFFSET];
		tmp_bwg_id = tmp_bwg_id >> 5;
		peer_pg_setting.tx.up[i].bandwidth_group_mapping = tmp_bwg_id;
		peer_pg_setting.rx.up[i].bandwidth_group_mapping = tmp_bwg_id;

		u8 tmp_strict_prio =
			port->portrx.rxmanifest->DCB_Type2->info_string
			[DCB_Type2_SETTINGS_OFFSET + 2*i + BYTE1_OFFSET];
		tmp_strict_prio = tmp_strict_prio >> 3;
		tmp_strict_prio &= 0x3;
		peer_pg_setting.tx.up[i].strict_priority =
			(dcb_strict_priority_type)tmp_strict_prio;
		peer_pg_setting.rx.up[i].strict_priority =
			(dcb_strict_priority_type)tmp_strict_prio;

		peer_pg_setting.tx.up[i].percent_of_bandwidth_group_cap =
			port->portrx.rxmanifest->DCB_Type2->info_string
			[DCB_Type2_SETTINGS_OFFSET + 2*i + BYTE2_OFFSET];
		peer_pg_setting.rx.up[i].percent_of_bandwidth_group_cap =
			port->portrx.rxmanifest->DCB_Type2->info_string
			[DCB_Type2_SETTINGS_OFFSET + 2*i + BYTE2_OFFSET];
		tmp_bwg_id = tmp_strict_prio = 0;

		peer_pg_setting.tx.up[i].traffic_class_mapping = 0;
		peer_pg_setting.rx.up[i].traffic_class_mapping = 0;
	}
	put_peer_pg(port->ifname, &peer_pg_setting);

	return(TRUE);
}

boolean_t process_dcb_type3_tlv(struct port *port)
{
	pfc_attribs  peer_pfc_setting;
	int i = 0;

	peer_pfc_setting.protocol.Advertise = TRUE;
	peer_pfc_setting.protocol.Oper_version =
		port->portrx.rxmanifest->DCB_Type3->info_string
		[DCB_HEADER_OPER_VERSION_OFFSET];
	peer_pfc_setting.protocol.Max_version =
		port->portrx.rxmanifest->DCB_Type3->info_string
		[DCB_HEADER_MAX_VERSION_OFFSET];
	if (port->portrx.rxmanifest->DCB_Type3->info_string
		[DCB_HEADER_EWE_OFFSET] & BIT7) {
		peer_pfc_setting.protocol.Enable = TRUE;
	} else {
		peer_pfc_setting.protocol.Enable = FALSE;
	}
	if (port->portrx.rxmanifest->DCB_Type3->info_string
		[DCB_HEADER_EWE_OFFSET] & BIT6) {
		peer_pfc_setting.protocol.Willing = TRUE;
	} else {
		peer_pfc_setting.protocol.Willing = FALSE;
	}
	if (port->portrx.rxmanifest->DCB_Type3->info_string
		[DCB_HEADER_EWE_OFFSET] & BIT5) {
		peer_pfc_setting.protocol.Error = TRUE;
	} else {
		peer_pfc_setting.protocol.Error = FALSE;
	}
	peer_pfc_setting.protocol.RxFeatureTLVPresent = TRUE;
	u8 temp = 0;
	for (i = 0; i < 8; i++) {
		temp = (u8)(port->portrx.rxmanifest->DCB_Type3->info_string
			[DCB_Type3_MAP_OFFSET]);
		peer_pfc_setting.admin[i] = (pfc_type)((temp >> i) & BIT0);
	}
	put_peer_pfc(port->ifname, &peer_pfc_setting);

	return(TRUE);
}


boolean_t process_dcb_type5_tlv(struct port *port)
{
	app_attribs  peer_app_setting;
	u32 i = 0;
	u8 sub_type = 0;

	peer_app_setting.protocol.Oper_version =
		port->portrx.rxmanifest->DCB_Type5->info_string
		[DCB_HEADER_OPER_VERSION_OFFSET];
	peer_app_setting.protocol.Max_version =
		port->portrx.rxmanifest->DCB_Type5->info_string
		[DCB_HEADER_MAX_VERSION_OFFSET];
	sub_type = port->portrx.rxmanifest->DCB_Type5->info_string
		[DCB_HEADER_SUB_TYPE_OFFSET];
	if (port->portrx.rxmanifest->DCB_Type5->info_string
		[DCB_HEADER_EWE_OFFSET] & BIT7) {
		peer_app_setting.protocol.Enable = TRUE;
	} else {
		peer_app_setting.protocol.Enable = FALSE;
	}
	if (port->portrx.rxmanifest->DCB_Type5->info_string
		[DCB_HEADER_EWE_OFFSET] & BIT6) {
		peer_app_setting.protocol.Willing = TRUE;
	} else {
		peer_app_setting.protocol.Willing = FALSE;
	}
	if (port->portrx.rxmanifest->DCB_Type5->info_string
		[DCB_HEADER_EWE_OFFSET] & BIT5) {
		peer_app_setting.protocol.Error = TRUE;
	} else {
		peer_app_setting.protocol.Error = FALSE;
	}
	peer_app_setting.protocol.RxFeatureTLVPresent = TRUE;
	u32 len = port->portrx.rxmanifest->DCB_Type5->length -
		sizeof(struct  dcb_tlv_header);
	peer_app_setting.Length = len;
	if (DCB_MAX_TLV_LENGTH < len) {
		return FALSE;
	}
	for (i = 0; i < len; i++) {
		peer_app_setting.AppData[i] =
			port->portrx.rxmanifest->DCB_Type5->info_string
			[DCB_Type5_APP_OFFSET + i];
	}
	put_peer_app(port->ifname, sub_type,&peer_app_setting);

	return(TRUE);
}

boolean_t process_dcb_type6_tlv(struct port *port)
{
	return(TRUE);
}

void    clear_manifest(struct port *port) {
	if (port->portrx.rxmanifest->DCB_Type6)
		port->portrx.rxmanifest->DCB_Type6 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->DCB_Type6);
	if (port->portrx.rxmanifest->DCB_Type5)
		port->portrx.rxmanifest->DCB_Type5 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->DCB_Type5);
	if (port->portrx.rxmanifest->DCB_Type4)
		port->portrx.rxmanifest->DCB_Type4 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->DCB_Type4);
	if (port->portrx.rxmanifest->DCB_Type3)
		port->portrx.rxmanifest->DCB_Type3 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->DCB_Type3);
	if (port->portrx.rxmanifest->DCB_Type2)
		port->portrx.rxmanifest->DCB_Type2 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->DCB_Type2);
	if (port->portrx.rxmanifest->DCB_Type1)
		port->portrx.rxmanifest->DCB_Type1 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->DCB_Type1);
	if (port->portrx.rxmanifest->Type127)
		port->portrx.rxmanifest->Type127 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->Type127);
	if (port->portrx.rxmanifest->Type8)
		port->portrx.rxmanifest->Type8 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->Type8);
	if (port->portrx.rxmanifest->Type7)
		port->portrx.rxmanifest->Type7 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->Type7);
	if (port->portrx.rxmanifest->Type6)
		port->portrx.rxmanifest->Type6 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->Type6);
	if (port->portrx.rxmanifest->Type5)
		port->portrx.rxmanifest->Type5 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->Type5);
	if (port->portrx.rxmanifest->Type4)
		port->portrx.rxmanifest->Type4 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->Type4);
	if (port->portrx.rxmanifest->Type3)
		port->portrx.rxmanifest->Type3 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->Type3);
	if (port->portrx.rxmanifest->Type2)
		port->portrx.rxmanifest->Type2 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->Type2);
	if (port->portrx.rxmanifest->Type1)
		port->portrx.rxmanifest->Type1 =
			destruct_unpacked_tlv(port->portrx.rxmanifest->Type1);
	free(port->portrx.rxmanifest);
	port->portrx.rxmanifest = NULL;
}
